@@ -1,40 +1,121 @@
-# mpris.py
-#
-# Copyright 2019 The GNOME Music developers
-# Copyright 2024 kramo
-#
-# This program is free software: you can redistribute it and/or modify
-# it under the terms of the GNU General Public License as published by
-# the Free Software Foundation, either version 3 of the License, or
-# (at your option) any later version.
-#
-# This program is distributed in the hope that it will be useful,
-# but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-# GNU General Public License for more details.
-#
-# You should have received a copy of the GNU General Public License
-# along with this program.  If not, see <http://www.gnu.org/licenses/>.
-#
 # SPDX-License-Identifier: GPL-3.0-or-later
+# SPDX-FileCopyrightText: 2019 The GNOME Music developers
+# SPDX-FileCopyrightText: Copyright 2024-2025 kramo
 
 # A lot of the code is taken from GNOME Music
 # https://gitlab.gnome.org/GNOME/gnome-music/-/blob/6a32efb74ff4107d1e4a288184e21c43f5dd877f/gnomemusic/mpris.py
 
-import logging
 import re
-from typing import Any, Optional
+from functools import partial
+from typing import Any
 
-from gi.repository import Gio, GLib, GstPlay  # type: ignore
+from gi.repository import (
+    Gio,
+    GLib,
+    GstAudio,  # pyright: ignore[reportAttributeAccessIssue]
+    GstPlay,  # pyright: ignore[reportAttributeAccessIssue]
+    Gtk,
+)
 
-from showtime import shared
+from showtime import APP_ID, PREFIX, logger
 from showtime.utils import get_title
-from showtime.window import ShowtimeWindow
+from showtime.widgets.window import Window
+
+RATE_SLOW = 0.75
+RATE_NORMAL = 1.125
+RATE_FASTER = 1.375
+RATE_FASTEST = 1.75
+
+INTERFACE = """
+<!DOCTYPE node PUBLIC
+'-//freedesktop//DTD D-BUS Object Introspection 1.0//EN'
+'http://www.freedesktop.org/standards/dbus/1.0/introspect.dtd'>
+<node>
+    <interface name='org.freedesktop.DBus.Introspectable'>
+        <method name='Introspect'>
+            <arg name='data' direction='out' type='s'/>
+        </method>
+    </interface>
+    <interface name='org.freedesktop.DBus.Properties'>
+        <method name='Get'>
+            <arg name='interface' direction='in' type='s'/>
+            <arg name='property' direction='in' type='s'/>
+            <arg name='value' direction='out' type='v'/>
+        </method>
+        <method name='Set'>
+            <arg name='interface_name' direction='in' type='s'/>
+            <arg name='property_name' direction='in' type='s'/>
+            <arg name='value' direction='in' type='v'/>
+        </method>
+        <method name='GetAll'>
+            <arg name='interface' direction='in' type='s'/>
+            <arg name='properties' direction='out' type='a{sv}'/>
+        </method>
+        <signal name='PropertiesChanged'>
+            <arg name='interface_name' type='s' />
+            <arg name='changed_properties' type='a{sv}' />
+            <arg name='invalidated_properties' type='as' />
+        </signal>
+    </interface>
+    <interface name='org.mpris.MediaPlayer2'>
+        <method name='Raise'>
+        </method>
+        <method name='Quit'>
+        </method>
+        <property name='CanQuit' type='b' access='read' />
+        <property name='Fullscreen' type='b' access='readwrite' />
+        <property name='CanRaise' type='b' access='read' />
+        <property name='HasTrackList' type='b' access='read'/>
+        <property name='Identity' type='s' access='read'/>
+        <property name='DesktopEntry' type='s' access='read'/>
+        <property name='SupportedUriSchemes' type='as' access='read'/>
+        <property name='SupportedMimeTypes' type='as' access='read'/>
+    </interface>
+    <interface name='org.mpris.MediaPlayer2.Player'>
+        <method name='Next'/>
+        <method name='Previous'/>
+        <method name='Pause'/>
+        <method name='PlayPause'/>
+        <method name='Stop'/>
+        <method name='Play'/>
+        <method name='Seek'>
+            <arg direction='in' name='Offset' type='x'/>
+        </method>
+        <method name='SetPosition'>
+            <arg direction='in' name='TrackId' type='o'/>
+            <arg direction='in' name='Position' type='x'/>
+        </method>
+        <method name='OpenUri'>
+            <arg direction='in' name='Uri' type='s'/>
+        </method>
+        <signal name='Seeked'>
+            <arg name='Position' type='x'/>
+        </signal>
+        <property name='PlaybackStatus' type='s' access='read'/>
+        <property name='LoopStatus' type='s' access='readwrite'/>
+        <property name='Rate' type='d' access='readwrite'/>
+        <property name='Shuffle' type='b' access='readwrite'/>
+        <property name='Metadata' type='a{sv}' access='read'>
+        </property>
+        <property name='Position' type='x' access='read'/>
+        <property name='MinimumRate' type='d' access='read'/>
+        <property name='MaximumRate' type='d' access='read'/>
+        <property name='CanGoNext' type='b' access='read'/>
+        <property name='CanGoPrevious' type='b' access='read'/>
+        <property name='CanPlay' type='b' access='read'/>
+        <property name='CanPause' type='b' access='read'/>
+        <property name='CanSeek' type='b' access='read'/>
+        <property name='CanControl' type='b' access='read'/>
+    </interface>
+</node>
+"""
 
 
 class DBusInterface:
+    """A D-Bus interface."""
+
     def __init__(self, name: str, path: str, _application: Any) -> None:
-        """Etablish a D-Bus session connection
+        """Etablish a D-Bus session connection.
 
         :param str name: interface name
         :param str path: object path
@@ -48,7 +129,7 @@ class DBusInterface:
         try:
             self._con = Gio.bus_get_finish(res)
         except GLib.Error as e:
-            logging.warning("Unable to connect to to session bus: %s", e.message)
+            logger.warning("Unable to connect to the session bus: %s", e.message)
             return
 
         Gio.bus_own_name_on_connection(
@@ -58,9 +139,7 @@ class DBusInterface:
         method_outargs = {}
         method_inargs = {}
         signals = {}
-        for interface in Gio.DBusNodeInfo.new_for_xml(
-            self.__doc__  #  type: ignore
-        ).interfaces:
+        for interface in Gio.DBusNodeInfo.new_for_xml(INTERFACE).interfaces:
             for method in interface.methods:
                 method_outargs[method.name] = (
                     "(" + "".join([arg.signature for arg in method.out_args]) + ")"
@@ -87,9 +166,9 @@ class DBusInterface:
 
     def _on_method_call(
         self,
-        connection: Gio.DBusConnection,
-        sender: str,
-        object_path: str,
+        _connection: Gio.DBusConnection,
+        _sender: str,
+        _object_path: str,
         interface_name: str,
         method_name: str,
         parameters: GLib.Variant,
@@ -110,7 +189,7 @@ class DBusInterface:
             if sig == "h":
                 msg = invocation.get_message()
                 fd_list = msg.get_unix_fd_list()
-                args[i] = fd_list.get(args[i])  #  type: ignore
+                args[i] = fd_list.get(args[i])  # pyright: ignore[reportOptionalMemberAccess]
 
         method_snake_name = DBusInterface.camelcase_to_snake_case(method_name)
         try:
@@ -149,111 +228,32 @@ class DBusInterface:
 
     @staticmethod
     def camelcase_to_snake_case(name: str) -> str:
+        """Convert `name` from camelCase to snake_case."""
         s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
         return "_" + re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()
 
 
 class MPRIS(DBusInterface):
-    """
-    <!DOCTYPE node PUBLIC
-    '-//freedesktop//DTD D-BUS Object Introspection 1.0//EN'
-    'http://www.freedesktop.org/standards/dbus/1.0/introspect.dtd'>
-    <node>
-        <interface name='org.freedesktop.DBus.Introspectable'>
-            <method name='Introspect'>
-                <arg name='data' direction='out' type='s'/>
-            </method>
-        </interface>
-        <interface name='org.freedesktop.DBus.Properties'>
-            <method name='Get'>
-                <arg name='interface' direction='in' type='s'/>
-                <arg name='property' direction='in' type='s'/>
-                <arg name='value' direction='out' type='v'/>
-            </method>
-            <method name='Set'>
-                <arg name='interface_name' direction='in' type='s'/>
-                <arg name='property_name' direction='in' type='s'/>
-                <arg name='value' direction='in' type='v'/>
-            </method>
-            <method name='GetAll'>
-                <arg name='interface' direction='in' type='s'/>
-                <arg name='properties' direction='out' type='a{sv}'/>
-            </method>
-            <signal name='PropertiesChanged'>
-                <arg name='interface_name' type='s' />
-                <arg name='changed_properties' type='a{sv}' />
-                <arg name='invalidated_properties' type='as' />
-            </signal>
-        </interface>
-        <interface name='org.mpris.MediaPlayer2'>
-            <method name='Raise'>
-            </method>
-            <method name='Quit'>
-            </method>
-            <property name='CanQuit' type='b' access='read' />
-            <property name='Fullscreen' type='b' access='readwrite' />
-            <property name='CanRaise' type='b' access='read' />
-            <property name='HasTrackList' type='b' access='read'/>
-            <property name='Identity' type='s' access='read'/>
-            <property name='DesktopEntry' type='s' access='read'/>
-            <property name='SupportedUriSchemes' type='as' access='read'/>
-            <property name='SupportedMimeTypes' type='as' access='read'/>
-        </interface>
-        <interface name='org.mpris.MediaPlayer2.Player'>
-            <method name='Next'/>
-            <method name='Previous'/>
-            <method name='Pause'/>
-            <method name='PlayPause'/>
-            <method name='Stop'/>
-            <method name='Play'/>
-            <method name='Seek'>
-                <arg direction='in' name='Offset' type='x'/>
-            </method>
-            <method name='SetPosition'>
-                <arg direction='in' name='TrackId' type='o'/>
-                <arg direction='in' name='Position' type='x'/>
-            </method>
-            <method name='OpenUri'>
-                <arg direction='in' name='Uri' type='s'/>
-            </method>
-            <signal name='Seeked'>
-                <arg name='Position' type='x'/>
-            </signal>
-            <property name='PlaybackStatus' type='s' access='read'/>
-            <property name='LoopStatus' type='s' access='readwrite'/>
-            <property name='Rate' type='d' access='readwrite'/>
-            <property name='Shuffle' type='b' access='readwrite'/>
-            <property name='Metadata' type='a{sv}' access='read'>
-            </property>
-            <property name='Position' type='x' access='read'/>
-            <property name='MinimumRate' type='d' access='read'/>
-            <property name='MaximumRate' type='d' access='read'/>
-            <property name='CanGoNext' type='b' access='read'/>
-            <property name='CanGoPrevious' type='b' access='read'/>
-            <property name='CanPlay' type='b' access='read'/>
-            <property name='CanPause' type='b' access='read'/>
-            <property name='CanSeek' type='b' access='read'/>
-            <property name='CanControl' type='b' access='read'/>
-        </interface>
-    </node>
-    """
+    """An MRPIS implementation."""
 
     MEDIA_PLAYER2_IFACE = "org.mpris.MediaPlayer2"
     MEDIA_PLAYER2_PLAYER_IFACE = "org.mpris.MediaPlayer2.Player"
 
     @property
-    def win(self) -> Optional[ShowtimeWindow]:
-        return self._app.get_active_window()
+    def win(self) -> Window | None:  # pyright: ignore[reportAttributeAccessIssue]
+        """Get the active application window."""
+        return win if isinstance(win := self._app.get_active_window(), Window) else None
 
     @property
-    def play(self) -> Optional[GstPlay.Play]:
+    def play(self) -> GstPlay.Play | None:
+        """Play the video."""
         if not self.win:
             return None
 
         return getattr(self.win, "play", None)
 
-    def __init__(self, app) -> None:
-        name = f"org.mpris.MediaPlayer2.{shared.APP_ID}"
+    def __init__(self, app: Gtk.Application) -> None:
+        name = f"org.mpris.MediaPlayer2.{APP_ID}"
         path = "/org/mpris/MediaPlayer2"
         super().__init__(name, path, app)
 
@@ -262,6 +262,9 @@ class MPRIS(DBusInterface):
         self._app.connect("state-changed", self._on_player_state_changed)
         self._app.connect("media-info-updated", self._on_media_info_updated)
         self._app.connect("notify::active-window", self._on_active_window_changed)
+        self._app.connect("volume-changed", self._on_volume_changed)
+        self._app.connect("rate-changed", self._on_rate_changed)
+        self._app.connect("seeked", self._on_seeked)
 
     def _get_playback_status(self) -> str:
         if (not self.win) or self.win.stopped:
@@ -275,23 +278,17 @@ class MPRIS(DBusInterface):
     def _get_metadata(self) -> dict:
         if (not self.play) or (not (media_info := self.play.get_media_info())):
             return {
-                "mpris:trackid": GLib.Variant(
-                    "o", f"{shared.PREFIX}/TrackList/CurrentTrack"
-                )
+                "mpris:trackid": GLib.Variant("o", f"{PREFIX}/TrackList/CurrentTrack")
             }
 
-        length = int(self.play.get_duration() / 1e6)
-
-        metadata = {
+        return {
             "xesam:url": GLib.Variant("s", media_info.get_uri()),
-            "mpris:length": GLib.Variant("x", length),
+            "mpris:length": GLib.Variant("x", int(self.play.get_duration() / 1e3)),
             "xesam:title": GLib.Variant(
                 "s", get_title(media_info) or _("Unknown Title")
             ),
         }
 
-        return metadata
-
     def _on_player_state_changed(self, *_args: Any) -> None:
         playback_status = self._get_playback_status()
 
@@ -315,7 +312,6 @@ class MPRIS(DBusInterface):
         )
 
     def _on_active_window_changed(self, *_args: Any) -> None:
-        position_msecond = int(self.play.get_position() / 1e6) if self.play else 0
         playback_status = self._get_playback_status()
         can_play = (self.play.get_uri() is not None) if self.play else False
         self._properties_changed(
@@ -323,13 +319,48 @@ class MPRIS(DBusInterface):
             {
                 "PlaybackStatus": GLib.Variant("s", playback_status),
                 "Metadata": GLib.Variant("a{sv}", self._get_metadata()),
-                "Position": GLib.Variant("x", position_msecond),
                 "CanPlay": GLib.Variant("b", can_play),
                 "CanPause": GLib.Variant("b", can_play),
             },
             [],
         )
 
+    def _on_volume_changed(self, *_args: Any) -> None:
+        if not self.win:
+            return
+
+        volume = self.win.pipeline.get_volume(GstAudio.StreamVolumeFormat.CUBIC)  # pyright: ignore[reportAttributeAccessIssue]
+
+        self._properties_changed(
+            MPRIS.MEDIA_PLAYER2_PLAYER_IFACE,
+            {
+                "Volume": GLib.Variant("d", volume),
+            },
+            [],
+        )
+
+    def _on_rate_changed(self, *_args: Any) -> None:
+        if not self.play:
+            return
+
+        self._properties_changed(
+            MPRIS.MEDIA_PLAYER2_PLAYER_IFACE,
+            {
+                "Rate": GLib.Variant("d", self.play.props.rate),
+            },
+            [],
+        )
+
+    def _on_seeked(self, *_args: Any) -> None:
+        position_usecond = int(self.play.get_position() / 1e3) if self.play else 0
+
+        self._dbus_emit_signal(
+            "Seeked",
+            {
+                "Position": position_usecond,
+            },
+        )
+
     def _raise(self) -> None:
         """Brings user interface to the front (MPRIS Method)."""
         if not self.win:
@@ -343,14 +374,12 @@ class MPRIS(DBusInterface):
 
     def _next(self) -> None:
         """Skips to the next track in the tracklist (MPRIS Method)."""
-        ...
 
     def _previous(self) -> None:
         """Skips to the previous track in the tracklist.
 
         (MPRIS Method)
         """
-        ...
 
     def _pause(self) -> None:
         """Pauses playback (MPRIS Method)."""
@@ -364,7 +393,7 @@ class MPRIS(DBusInterface):
         if not self.win:
             return
 
-        self.win.toggle_playback()
+        self.win.unpause() if self.win.paused else self.win.pause()
 
     def _stop(self) -> None:
         """Stop playback (MPRIS Method)."""
@@ -384,83 +413,78 @@ class MPRIS(DBusInterface):
 
         self.win.unpause()
 
-    def _seek(self, offset_msecond: int) -> None:
+    def _seek(self, offset_usecond: int) -> None:
         """Seek forward in the current track (MPRIS Method).
 
         Seek is relative to the current player position.
         If the value passed in would mean seeking beyond the end of the track,
         acts like a call to Next.
 
-        :param int offset_msecond: number of microseconds
+        :param int offset_usecond: number of microseconds
         """
         if not self.play:
             return
 
-        self.play.seek(max(0, self.play.get_position() + (offset_msecond * 1e6)))
+        self.play.seek(max(0, self.play.get_position() + (offset_usecond * 1e3)))
 
-    def _set_position(self, _track_id: str, position_msecond: int) -> None:
-        """Set the current track position in microseconds (MPRIS Method)
+    def _set_position(self, _track_id: str, position_usecond: int) -> None:
+        """Set the current track position in microseconds (MPRIS Method).
 
         :param str track_id: The currently playing track's identifier
-        :param int position_msecond: new position in microseconds
+        :param int position_usecond: new position in microseconds
         """
         if not self.play:
             return
 
-        self.play.seek(position_msecond * 1e6)
+        self.play.seek(position_usecond * 1e3)
 
     def _open_uri(self, _uri: str) -> None:
-        """Opens the Uri given as an argument (MPRIS Method).
+        """Open the Uri given as an argument (MPRIS Method).
 
         Not implemented.
 
         :param str uri: Uri of the track to load.
         """
-        ...
 
-    def _get(self, interface_name: str, property_name: str) -> dict:
+    def _get(self, iface: str, prop: str) -> GLib.Variant | None:
         # Some clients (for example GSConnect) try to access the volume
         # property. This results in a crash at startup.
         # Return nothing to prevent it.
         try:
-            return self._get_all(interface_name)[property_name]  # type: ignore
-        except KeyError:
-            msg = "MPRIS does not handle {} property from {} interface".format(
-                property_name, interface_name
-            )
-            logging.warning(msg)
-            raise ValueError(msg)
+            return all_props[prop] if (all_props := self._get_all(iface)) else None
+        except KeyError as error:
+            msg = f"MPRIS does not handle {prop} property from {iface} interface"
+            logger.warning(msg)
+            raise ValueError(msg) from error
 
-    def _get_all(self, interface_name: str) -> Optional[dict]:
+    def _get_all(self, interface_name: str) -> dict | None:
         if interface_name == MPRIS.MEDIA_PLAYER2_IFACE:
-            application_id = self._app.props.application_id
             return {
                 "CanQuit": GLib.Variant("b", True),
                 "Fullscreen": GLib.Variant("b", False),
                 "CanSetFullscreen": GLib.Variant("b", False),
                 "CanRaise": GLib.Variant("b", False),
                 "HasTrackList": GLib.Variant("b", False),
-                "Identity": GLib.Variant("s", "Showtime"),
-                "DesktopEntry": GLib.Variant("s", shared.APP_ID),
+                "Identity": GLib.Variant("s", "Video Player"),
+                "DesktopEntry": GLib.Variant("s", APP_ID),
                 "SupportedUriSchemes": GLib.Variant("as", ["file"]),
-                "SupportedMimeTypes": GLib.Variant(
-                    "as",
-                    [],
-                ),
+                "SupportedMimeTypes": GLib.Variant("as", []),
             }
-        elif interface_name == MPRIS.MEDIA_PLAYER2_PLAYER_IFACE:
-            position_msecond = int(self.play.get_position() / 1e6) if self.play else 0
-            playback_status = self._get_playback_status()
+
+        if interface_name == MPRIS.MEDIA_PLAYER2_PLAYER_IFACE:
+            position_usecond = int(self.play.get_position() / 1e3) if self.play else 0
+            volume = self.play.get_volume() if self.play else 0.0
             can_play = (self.play.get_uri() is not None) if self.play else False
             return {
-                "PlaybackStatus": GLib.Variant("s", playback_status),
+                "PlaybackStatus": GLib.Variant("s", self._get_playback_status()),
                 "LoopStatus": GLib.Variant("s", "None"),
-                "Rate": GLib.Variant("d", 1.0),
+                "Rate": GLib.Variant("d", self.play.props.rate if self.play else 0.0),
                 "Shuffle": GLib.Variant("b", False),
                 "Metadata": GLib.Variant("a{sv}", self._get_metadata()),
-                "Position": GLib.Variant("x", position_msecond),
-                "MinimumRate": GLib.Variant("d", 1.0),
-                "MaximumRate": GLib.Variant("d", 1.0),
+                "Volume": GLib.Variant("d", volume),
+                "Position": GLib.Variant("x", position_usecond),
+                "MinimumRate": GLib.Variant("d", 0.5),
+                "MaximumRate": GLib.Variant("d", 2.0),
                 "CanGoNext": GLib.Variant("b", False),
                 "CanGoPrevious": GLib.Variant("b", False),
                 "CanPlay": GLib.Variant("b", can_play),
@@ -468,26 +492,49 @@ class MPRIS(DBusInterface):
                 "CanSeek": GLib.Variant("b", True),
                 "CanControl": GLib.Variant("b", True),
             }
-        elif interface_name == "org.freedesktop.DBus.Properties":
-            return {}
-        elif interface_name == "org.freedesktop.DBus.Introspectable":
-            return {}
-        else:
-            logging.warning("MPRIS does not implement %s interface", interface_name)
 
-    def _set(self, interface_name: str, property_name: str, _new_value: Any) -> None:
-        if interface_name == MPRIS.MEDIA_PLAYER2_IFACE:
-            if property_name == "Fullscreen":
+        if interface_name in (
+            "org.freedesktop.DBus.Properties",
+            "org.freedesktop.DBus.Introspectable",
+        ):
+            return {}
+
+        logger.warning("MPRIS does not implement %s interface", interface_name)
+        return None
+
+    def _set(self, interface_name: str, property_name: str, value: Any) -> None:
+        if interface_name != MPRIS.MEDIA_PLAYER2_PLAYER_IFACE:
+            logger.warning("MPRIS does not implement %s interface", interface_name)
+            return
+
+        if not self.win:
+            return
+
+        match property_name:
+            case "Rate":
+                self.win.rate = (
+                    "0.5"
+                    if value < RATE_SLOW
+                    else "1.0"
+                    if value < RATE_NORMAL
+                    else "1.25"
+                    if value < RATE_FASTER
+                    else "1.5"
+                    if value < RATE_FASTEST
+                    else "2.0"
+                )
+            case "Volume":
+                GLib.idle_add(
+                    partial(
+                        self.win.pipeline.set_volume,  # pyright: ignore[reportAttributeAccessIssue]
+                        GstAudio.StreamVolumeFormat.CUBIC,
+                        value,
+                    )
+                )
+            case "LoopStatus":
                 pass
-        elif interface_name == MPRIS.MEDIA_PLAYER2_PLAYER_IFACE:
-            if property_name in ["Rate", "Volume"]:
+            case "Shuffle":
                 pass
-            elif property_name == "LoopStatus":
-                pass
-            elif property_name == "Shuffle":
-                pass
-        else:
-            logging.warning("MPRIS does not implement %s interface", interface_name)
 
     def _properties_changed(
         self,
@@ -495,12 +542,14 @@ class MPRIS(DBusInterface):
         changed_properties: dict,
         invalidated_properties: list,
     ) -> None:
-        parameters = {
-            "interface_name": interface_name,
-            "changed_properties": changed_properties,
-            "invalidated_properties": invalidated_properties,
-        }
-        self._dbus_emit_signal("PropertiesChanged", parameters)
+        self._dbus_emit_signal(
+            "PropertiesChanged",
+            {
+                "interface_name": interface_name,
+                "changed_properties": changed_properties,
+                "invalidated_properties": invalidated_properties,
+            },
+        )
 
-    def _introspect(self) -> Optional[str]:
-        return self.__doc__
+    def _introspect(self) -> str | None:
+        return INTERFACE
